from __future__ import annotations

import joblib

import numpy as np
from scipy.signal import fftconvolve
from sklearn.decomposition import IncrementalPCA, PCA
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.base import OutlierMixin
from sklearn.pipeline import Pipeline
from sklearn.kernel_approximation import Nystroem
from sklearn.linear_model import BayesianRidge
from sklearn.metrics import accuracy_score
from scipy.stats import gaussian_kde
from itertools import product

from sklearn.base import clone, MetaEstimatorMixin
from joblib import Parallel, delayed
from copy import deepcopy

from pes_to_spec.bnn import BNNModel

from typing import Any, Dict, List, Optional, Union, Tuple

def matching_ids(a: np.ndarray, b: np.ndarray, c: np.ndarray) -> np.ndarray:
    """Returns list of train IDs common to sets a, b and c."""
    unique_ids = list(set(a).intersection(b).intersection(c))
    return np.array(unique_ids)

def matching_two_ids(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Returns list of train IDs common to sets a and b."""
    unique_ids = list(set(a).intersection(b))
    return np.array(unique_ids)

class PromptNotFoundError(Exception):
    """
    Exception representing the error condition generated by not finding the prompt peak.
    """
    def __init__(self):
        pass
    def __str__(self) -> str:
        return "No prompt peak has been detected."


class HighResolutionSmoother(TransformerMixin, BaseEstimator):
    """
    Smoothens out the high resolution data.

    Args:
      high_res_sigma: Energy resolution in eV.
    """
    def __init__(self,
                 high_res_sigma: float=0.2
                 ):
        self.high_res_sigma = high_res_sigma
        self.energy = None

    def fit(self, X, y=None, **fit_params) -> TransformerMixin:
        """
        Fit records the energy axis.

        Args:
          X: Irrelevant.
          y: Irrelevant.
          fit_params: Contains the energy axis in the key "energy" with shape (any, features).

        Returns: The object itself.
        """
        self.energy = np.copy(fit_params["energy"])
        if len(self.energy.shape) == 2:
            self.energy = self.energy[0,:]
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Apply smoothing in X using the energy axis.

        Args:
          X: Input to smoothen with shape (train id, features).

        Returns: Smoothened out spectrum.
        """
        if self.high_res_sigma <= 0:
            return X
        # use a default energy axis is none is given
        # assume only the energy step
        energy = np.broadcast_to(self.energy, X.shape)

        # Apply smoothing
        n_features = X.shape[1]
        # get the centre value of the energy axis
        mu = energy[:, n_features//2, np.newaxis]
        # generate a gaussian
        gaussian = np.exp(-0.5*(energy - mu)**2/self.high_res_sigma**2)
        gaussian /= np.sum(gaussian, axis=1, keepdims=True)
        # apply it to the data
        high_res_gc = fftconvolve(X, gaussian, mode="same", axes=1)
        return high_res_gc

    def inverse_transform(self, Xt: np.ndarray) -> np.ndarray:
        """
        Dummy function to allow for an inverse transform.

        Args:
          Xt: Transformed variable.

        Returns: Xt.
        """
        return Xt

class UncertaintyHolder(TransformerMixin, BaseEstimator):
    """
    Keep track of uncertainties.

    """
    def __init__(self):
        self.unc: np.ndarray = np.zeros((1, 0), dtype=float)

    def set_uncertainty(self, unc: np.ndarray):
        """
        Set the uncertainty.

        Args:
          unc: The uncertainty.
        """
        self.unc = np.copy(unc)

    def fit(self, X, y=None) -> TransformerMixin:
        """
        Does nothing.

        Args:
          X: Irrelevant.
          y: Irrelevant.

        Returns: Itself.
        """
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

    def uncertainty(self):
        """The uncertainty recorded."""
        return self.unc

class DataHolder(TransformerMixin, BaseEstimator):
    """
    Keep track of relevance dictionaries.

    """
    def __init__(self, data: Dict[str, Any]=dict()):
        self.data: Dict[str, Any] = data

    def set_data(self, data: Dict[str, Any]):
        """
        Set.

        Args:
          data: Data.
        """
        self.data = deepcopy(data)

    def get_data(self) -> Dict[str, Any]:
        """
        Get.

        """
        return self.data

    def fit(self, X, y=None) -> TransformerMixin:
        """
        Does nothing.

        Args:
          X: Irrelevant.
          y: Irrelevant.

        Returns: Itself.
        """
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

class SelectRelevantLowResolution(TransformerMixin, BaseEstimator):
    """
    Select only relevant entries in the low-resolution data.

    Args:
      channels: Selected channels to use as an input for the low resolution data.
      tof_start: Start looking at this index from the low-resolution spectrometer data.
                 Set to None to perform no selection
      delta_tof: Number of components to take from the low-resolution spectrometer.
                 Set to None to perform no selection.
      poly: Whether to output a polynomial expantion of the low-resolution data.
    """
    def __init__(self,
                 channels:List[str]=[f"channel_{j}_{k}"
                                     for j, k in product(range(1, 5), ["A", "B", "C", "D"])],
                 tof_start: Optional[int]=None,
                 delta_tof: Optional[int]=300,
                 poly: bool=False
                 ):
        self.channels = channels
        self.tof_start = tof_start
        self.delta_tof = delta_tof
        self.poly = poly
        self.mean = dict()
        self.std = dict()

    def transform(self, X: Dict[str, np.ndarray],
                  keep_dictionary_structure: bool=False,
                  pulse_spacing: Optional[Dict[str, List[int]]]=None,
                  pulse_energy: Optional[np.ndarray]=None) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """
        Get a dictionary with the channel names for the inut low resolution data and output
        only the relevant input data in an array.

        Args:
          X: Dictionary with keys named channel_{i}_{k},
             where i is a number between 1 and 4 and k is a letter between A and D.
          keep_dictionary_structure: Whether to concatenate all channels, or keep them as a dictionary.
          pulse_spacing: Distances between pulses in multi-pulse data. If there is only one pulse, set it to a list containing only the element zero.
          pulse_energy: Pulse energy.

        Returns: Concatenated and pre-processed low-resolution data of shape (train_id, pulse_id, features).
        """
        if self.tof_start is None:
            raise NotImplementedError("The low-resolution data cannot be transformed before the prompt has been identified. Call the fit function first.")
        if pulse_spacing is None:
            pulse_spacing = {ch: [0] for ch in X.keys()}
        y = X
        if self.delta_tof is not None:
            first = max(0, self.tof_start - self.delta_tof)
            last = min(X[self.channels[0]].shape[1], self.tof_start + self.delta_tof)
            y = {channel: np.stack([item[:, (first + delta):(last + delta)] for delta in pulse_spacing[channel]], axis=1)
                 for channel, item in X.items()}
        if not keep_dictionary_structure:
            selected = list(y.values())
            if pulse_energy is not None:
                selected += [pulse_energy[:, np.newaxis, :]]
                if self.poly:
                    selected += [pulse_energy[:, np.newaxis, :]*v for v in y.values()]
            return np.concatenate(selected, axis=-1)
        return y

    def estimate_prompt_peak(self, X: Dict[str, np.ndarray]) -> int:
        """
        Estimate the prompt peak index.

        Args:
          X: Low resolution data with a dictionary containing the channel names.

        Returns: The index.
        """
        # reduce on channel and on train ID
        sum_low_res = - np.mean(sum(list(X.values())), axis=0)
        axis = np.arange(0.0, sum_low_res.shape[0], 1.0)
        #widths = np.arange(10, 50, step=5)
        #peak_idx = find_peaks_cwt(sum_low_res, widths)
        gaussian = np.exp(-0.5*(axis - sum_low_res.shape[0]//2)**2/20**2)
        gaussian /= np.sum(gaussian, axis=0, keepdims=True)
        # apply it to the data
        smoothened = fftconvolve(sum_low_res, gaussian, mode="same", axes=0)
        peak_idx = [np.argmax(smoothened)]
        if len(peak_idx) < 1:
            raise PromptNotFoundError()
        peak_idx = sorted(peak_idx, key=lambda k: np.fabs(sum_low_res[k]), reverse=True)
        best_guess = int(peak_idx[0])
        # look around this estimate for the maximum
        # this is probably not necessary
        #min_search = max(best_guess - 10, 0)
        #max_search = min(best_guess + 10, len(sum_low_res))
        #restricted_arr = sum_low_res[min_search:max_search]
        #improved_guess = min_search + int(np.argmax(restricted_arr))
        return best_guess

    def fit(self, X: Dict[str, np.ndarray], y: Optional[Any]=None) -> TransformerMixin:
        """
        Estimate the prompt peak index.

        Args:
          X: Low resolution data with a dictionary containing the channel names.
          y: Ignored.

        Returns: The object itself.
        """
        self.tof_start = self.estimate_prompt_peak(X)
        X_tr = self.transform(X, keep_dictionary_structure=True)
        self.mean = {ch: np.mean(X_tr[ch], axis=0, keepdims=True)
                     for ch in X.keys()}
        self.std = {ch: np.std(X_tr[ch], axis=0, keepdims=True)
                    for ch in X.keys()}
        return self

    def debug_peak_finding(self, X: Dict[str, np.ndarray], filename: str):
        """
        Produce image to understand if the peak finding step worked well.

        Args:
          X: Low resolution data with a dictionary containing the channel names.
          filename: The file name where to save the plot.

        """
        sum_low_res = - np.mean(sum(list(X.values())), axis=0)
        peak_idx = self.estimate_prompt_peak(X)
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(8, 16))
        ax = plt.gca()
        ax.plot(np.arange(peak_idx-300, peak_idx+300),
                sum_low_res[peak_idx-300:peak_idx+300],
                c="b",
                label="Data")
        ax.set(title="",
               xlabel="Photon Spectrometer channel",
               ylabel="Sum of all Photon Spectrometer channels")
        plt.axvline(peak_idx,
                linewidth=3,
                ls="--",
                color='r',
                label="Peak position")
        ax.legend()
        plt.savefig(filename)
        plt.close(fig)

def _fit_estimator(estimator, X: np.ndarray, y: np.ndarray, w: np.ndarray):
    estimator = clone(estimator)
    estimator.fit(X, y, w)
    return estimator

class MultiOutputWithStd(MetaEstimatorMixin, BaseEstimator):

    def __init__(self, estimator, *, n_jobs=-1):
        self.estimator = estimator
        self.n_jobs = n_jobs

    def fit(self, X: np.ndarray, y: np.ndarray, weights: Optional[np.ndarray]=None):
        """Fit the model to data, separately for each output variable.

        Args:
          X: {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.
          y: {array-like, sparse matrix} of shape (n_samples, n_outputs)
            Multi-output targets. An indicator matrix turns on multilabel
            estimation.

        Returns: self.
        """
        if y.ndim == 1:
            raise ValueError(
                "y must have at least two dimensions for "
                "multi-output regression but has only one."
            )
        if weights is None:
            weights = np.ones(y.shape[0])

        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_estimator)(
                self.estimator, X, y[:, i], weights
            )
            for i in range(y.shape[1])
        )
        self.fast_coef = np.concatenate([e.coef_.reshape(-1, 1) for e in self.estimators_], axis=1)
        self.fast_intercept = np.array([e.intercept_ for e in self.estimators_]).reshape(1, -1)
        self.fast_inv_alpha = np.array([1.0/e.alpha_ for e in self.estimators_]).reshape(1, -1)
        self.fast_sigma = np.stack([e.sigma_ for e in self.estimators_], axis=0)

        return self

    def predict(self, X: np.ndarray, return_std: bool=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Predict multi-output variable using model for each target variable.

        Args:
          X: {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        Returns: {array-like, sparse matrix} of shape (n_samples, n_outputs)
            Multi-output targets predicted across multiple predictors.
            Note: Separate models are generated for each predictor.
        """
        y = X @ self.fast_coef + self.fast_intercept
        if not return_std:
            return y
        sigmas_squared_data = ((X @ self.fast_sigma) * X).sum(axis=-1).T
        y_std = np.sqrt(sigmas_squared_data + self.fast_inv_alpha)
        return y, y_std

class UncorrelatedDeviation(OutlierMixin, BaseEstimator):
    """
    Detect outliers from uncorrelated inputs.
    It uses a chi^2 sum over the features to flatten the features.
    The standard deviation is estimated using quantiles.

    Args:
      sigma: Number of standard deviations of the chi^2 distribution.

    """
    def __init__(self, sigma: float=5.0):
        super().__init__()
        self.sigma = sigma

    def fit(self, X, y=None) -> OutlierMixin:
        """
        Does nothing.

        Args:
          X: Data where outliers are.
          y: Irrelevant.

        Returns: Itself.
        """
        bounds_ = np.quantile(X, (0.003/2.0, 0.5, 1.0 - 0.003/2.0), axis=0)
        self.ndof_ = float(X.shape[1] - 1.0)
        self.med_ = bounds_[1, np.newaxis, ...]
        self.sigma_ = (bounds_[2, np.newaxis, ...] - bounds_[0, np.newaxis, ...])/3.0
        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Return the decision function.
        This is chi^2/ndof - 1 - sigma*sqrt(var_chi2)
        """
        return (self.score_samples(X) - 1.0 - self.sigma*np.sqrt(2.0/self.ndof_))

    def score_samples(self, X: np.ndarray) -> np.ndarray:
        """
        Return the Mahalanobis distance.

        Args:
          X: The new input data.

        Returns: The chi^2 test statistic.
        """
        return np.sum(((X - self.med_)/self.sigma_)**2, axis=1)/float(self.ndof_)

    def predict(self, X):
        """
        Predict labels (1 inlier, -1 outlier) of X according to fitted model.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix.
        Returns
        -------
        is_inlier : ndarray of shape (n_samples,)
            Returns -1 for anomalies/outliers and +1 for inliers.
        """
        values = self.decision_function(X)
        is_inlier = np.full(values.shape[0], 1, dtype=int)
        is_inlier[values > 0] = -1
        return is_inlier

    def score(self, X, y, sample_weight=None):
        """Return the mean accuracy on the given test data and labels.
        In multi-label classification, this is the subset accuracy
        which is a harsh metric since you require for each sample that
        each label set be correctly predicted.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            True labels for X.
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) w.r.t. y.
        """
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)

class Model(TransformerMixin, BaseEstimator):
    """
    Object representing a previous fit of the model to be used to predict high-resolution
    spectrum from a low-resolution one.

    Args:
      channels: Selected channels to use as an input for the low resolution data.
      n_pca_lr: Number of low-resolution data PCA components.
      n_pca_hr: Number of high-resolution data PCA components.
      high_res_sigma: Resolution of the high-resolution spectrometer in electron-Volts.
      tof_start: Start looking at this index from the low-resolution spectrometer data.
                 Set to None to perform no selection
      delta_tof: Number of components to take from the low-resolution spectrometer.
                 Set to None to perform no selection.
      validation_size: Fraction (number between 0 and 1) of the data to take for
                       validation and systematic uncertainty estimate.
      bnn: Use BNN?

    """
    def __init__(self,
                 channels:List[str]=[f"channel_{j}_{k}"
                                     for j, k in product(range(1, 5), ["A", "B", "C", "D"])],
                 n_pca_lr: int=600,
                 n_pca_hr: int=20,
                 high_res_sigma: float=0.2,
                 tof_start: Optional[int]=None,
                 delta_tof: Optional[int]=300,
                 validation_size: float=0.05,
                 bnn: bool=True,
                ):
        self.high_res_sigma = high_res_sigma
        # models
        self.x_select = SelectRelevantLowResolution(channels, tof_start, delta_tof, poly=not bnn)
        x_model_steps = list()
        x_model_steps += [
                          ('pca', PCA(n_pca_lr, whiten=True)),
                          ('unc', UncertaintyHolder()),
                          ]
        self.x_model = Pipeline(x_model_steps)
        self.y_model = Pipeline([
                                ('smoothen', HighResolutionSmoother(high_res_sigma)),
                                ('pca', PCA(n_pca_hr, whiten=True)),
                                ('unc', UncertaintyHolder()),
                                ])
        self.ood = {ch: UncorrelatedDeviation(sigma=5)
                    for ch in channels+['full']}
        if bnn:
            self.fit_model = BNNModel()
        else:
            self.fit_model = MultiOutputWithStd(BayesianRidge(n_iter=300, tol=1e-8, verbose=True), n_jobs=8)
        self.bnn = bnn

        self.kde_xgm = None
        self.mu_xgm = np.nan
        self.sigma_xgm = np.nan

        # we are reducing per channel from 2*delta_tof to delta_tof to check correlations
        n_pca_lr_per_channel = delta_tof
        self.channel_pca = {ch: IncrementalPCA(n_pca_lr_per_channel, whiten=True)
                            for ch in channels}

        # size of the test subset
        self.validation_size = validation_size

    def get_channels(self) -> List[str]:
        """Get channels used in training."""
        return self.x_select.channels

    def get_energy_values(self) -> np.ndarray:
        """Get x-axis of high-resolution data."""
        return self.y_model.named_steps["smoothen"].energy

    def get_low_resolution_range(self) -> Tuple[int, int]:
        """Get bin number with first and last relevant bins in the low-resolution spectrum."""
        first = (self.x_select.tof_start - self.x_select.delta_tof)
        last = (self.x_select.tof_start + self.x_select.delta_tof)
        return first, last

    def debug_peak_finding(self, low_res_data: Dict[str, np.ndarray], filename: str):
        """
        Produce image to understand if the peak finding step worked well.

        Args:
          low_res_data: Low resolution data with a dictionary containing the channel names.
          filename: The file name where to save the plot.

        """
        self.x_select.debug_peak_finding(low_res_data, filename)

    def preprocess_high_res(self, high_res_data: np.ndarray) -> np.ndarray:
        """
        Preprocess high-resolution data to remove high requency components.

        Args:
          high_res_data: The high-resolution data.

        Returns: Smoothened spectrum.
        """
        return self.y_model['smoothen'].transform(high_res_data)

    def uniformize(self, intensity: np.ndarray) -> np.ndarray:
        """
        Calculate weights to uniformize data in variable intensity.

        Args:
          intensity: The variable to uniformize the weights on.

        Return: weights.
        """
        self.kde_xgm = gaussian_kde(intensity.reshape(-1), bw_method="scott")
        self.mu_xgm = np.mean(intensity.reshape(-1), axis=0)
        self.sigma_xgm = np.std(intensity.reshape(-1), axis=0)
        q = np.quantile(intensity, [0.10, 0.90])
        l, h = q[0], q[1]
        x = intensity*((intensity > l) & (intensity < h)) + l*(intensity <= l) + h*(intensity >= h)
        log_prob = self.kde_xgm.logpdf(x.reshape(-1))
        w = np.exp(-log_prob)
        w = w/np.median(w)
        return w

    def fit(self, low_res_data: Dict[str, np.ndarray],
            high_res_data: np.ndarray, high_res_photon_energy: np.ndarray,
            weights: Optional[np.ndarray]=None,
            pulse_energy: Optional[np.ndarray]=None,
            ) -> np.ndarray:
        """
        Train the model.

        Args:
          low_res_data: Low resolution data as a dictionary with the key set to `channel_{i}_{k}`,
                        where i is a number between 1 and 4 and k is a letter between A and D.
                        For each dictionary entry, a numpy array is expected with shape
                        (train_id, ToF channel).
          high_res_data: Reference high resolution data with a one-to-one match to the
                         low resolution data in the train_id dimension. Shape (train_id, ToF channel).
          high_res_photon_energy: Photon energy axis for the high-resolution data.

        Returns: Smoothened high resolution spectrum.
        """
        if weights is None:
            weights = np.ones(high_res_data.shape[0])
        print("Fitting PCA on low-resolution data.")
        self.x_select.fit(low_res_data)
        low_res_select = self.x_select.transform(low_res_data, pulse_energy=pulse_energy)
        # keep the number of pulses
        B, P, _ = low_res_select.shape
        low_res_select = low_res_select.reshape((B*P, -1))

        n_components = min(self.x_model["pca"].n_components, low_res_select.shape[0])
        self.x_model.set_params(pca__n_components=n_components)
        x_t = self.x_model.fit_transform(low_res_select)
        print("Fitting PCA on high-resolution data.")
        y_t = self.y_model.fit_transform(high_res_data, smoothen__energy=high_res_photon_energy)

        print("Fitting outlier detection")
        self.ood['full'].fit(x_t)
        inliers = self.ood['full'].predict(x_t) > 0.0
        print("Fitting model.")
        self.fit_model.fit(x_t[inliers], y_t[inliers], weights[inliers])

        # calculate the effect of the PCA
        print("Calculate PCA unc. on high-resolution data.")
        high_res = self.y_model['smoothen'].transform(high_res_data)
        high_pca = self.y_model['pca'].transform(high_res)
        high_pca_rec = self.y_model['pca'].inverse_transform(high_pca)
        high_pca_unc = np.sqrt(np.mean((high_res - high_pca_rec)**2, axis=0, keepdims=True))
        self.y_model['unc'].set_uncertainty(high_pca_unc)

        print("Calculate transfer function")
        # Model: z(e) = conv(h(e), d(e)) + n(e)
        # d: true high-resolution data
        # h: the effect of the model
        # z: estimated high-resolution data
        # n: noise (uncertainty)
        # e: energy
        # true signal (as far as we can get -- it is smoothened, but this is the model target)
        d = high_res[inliers]
        D = np.fft.fft(d)

        y_pred, n = self.fit_model.predict(x_t[inliers], return_std=True)
        z = self.y_model['pca'].inverse_transform(y_pred)

        n = np.sqrt((self.y_model['pca'].inverse_transform(y_pred + n) - z)**2 + high_pca_unc**2)
        e = high_res_photon_energy[0,:] if len(high_res_photon_energy.shape) == 2 else high_res_photon_energy

        Z = np.fft.fft(z)
        V = np.fft.fft(np.mean(n, axis=0))

        de = e[1] - e[0]
        E = np.fft.fftfreq(len(e), de)
        e_axis = np.linspace(-0.5*(e[-1] - e[0]), 0.5*(e[-1] - e[0]), len(e))
        # generate a gaussian
        #gaussian = np.exp(-0.5*(e_axis)**2/self.high_res_sigma**2)
        #gaussian /= np.sum(gaussian, axis=0, keepdims=True)
        #gaussian = np.clip(gaussian, a_min=1e-6, a_max=None)
        #gaussian_ft = np.fft.fft(gaussian)

        H = np.mean(Z/D, axis=0)
        N = np.absolute(V)**2
        S = np.mean(np.absolute(D)**2, axis=0)
        H2 = np.absolute(H)**2
        nonzero = np.absolute(H) > 0.2
        Hinv = (1.0/H)*nonzero
        # Wiener filter:
        G = Hinv * (H2 * S) / (H2 * S + N)

        #import matplotlib.pyplot as plt
        #from matplotlib.gridspec import GridSpec
        #fig = plt.figure(figsize=(40, 40))
        #gs = GridSpec(2, 2)
        #ax = fig.add_subplot(gs[0, 0])
        #ax.plot(np.fft.fftshift(np.mean(np.absolute(Z), axis=0)), c='b', lw=3, label="Prediction")
        #ax.plot(np.fft.fftshift(np.mean(np.absolute(D), axis=0)), c='r', lw=3, label="True")
        #ax.legend()
        #ax.set(title=f"",
        #       xlabel="Reciprocal energy [1/eV]",
        #       ylabel="Intensity [a.u.]",
        #       yscale='log',
        #       )
        #ax = fig.add_subplot(gs[0, 1])
        #ax.plot(np.fft.fftshift(np.absolute(H)**2), c='b', lw=3, label=r"$|H|^2$")
        #ax.plot(np.fft.fftshift(np.absolute(Hinv)), c='k', lw=3, label=r"$|H^{-1}|$")
        #ax.plot(np.fft.fftshift(N), c='g', lw=3, label=r"$N$")
        #ax.plot(np.fft.fftshift(S), c='r', lw=3, label=r"$S$")
        #ax.legend()
        #ax.set(title=f"",
        #       xlabel="Reciprocal energy [1/eV]",
        #       ylabel="Intensity [a.u.]",
        #       yscale='log',
        #       )
        #ax = fig.add_subplot(gs[1, 0])
        #ax.plot(np.fft.fftshift(np.absolute(H)), c='b', lw=3, label="H")
        #ax.plot(np.fft.fftshift(np.absolute(np.mean(Z, axis=0)/np.mean(D, axis=0))), c='r', lw=3, label="mean Z/mean D")
        #ax.legend()
        #ax.set(title=f"",
        #       xlabel="Reciprocal energy [1/eV]",
        #       ylabel="Intensity [a.u.]",
        #       yscale="log",
        #       )
        #ax = fig.add_subplot(gs[1, 1])
        #ax.plot(np.fft.fftshift(np.absolute(G)), c='b', lw=3, label="H")
        #ax.plot(np.fft.fftshift(np.absolute(np.mean(Z, axis=0)/np.mean(D, axis=0))), c='r', lw=3, label="mean Z/mean D")
        #ax.legend()
        #ax.set(title=f"",
        #       xlabel="Reciprocal energy [1/eV]",
        #       ylabel="Intensity [a.u.]",
        #       yscale="log",
        #       )
        #fig.savefig("tmp.png")
        #plt.close(fig)

        #Hmod = np.real(np.absolute(H))
        Gdir = np.fft.fftshift(np.fft.ifft(G))
        self.wiener_filter = Gdir
        self.wiener_filter_ft = G
        self.wiener_energy = e_axis
        self.wiener_energy_ft = E
        self.transfer_function = H
        h = np.fft.fftshift(np.fft.ifft(H))
        hmod = np.real(np.absolute(h))
        self.impulse_response = h
        energy_mu = np.sum(e_axis*hmod)/np.sum(hmod)
        energy_var = np.sum(((e_axis - energy_mu)**2)*hmod)/np.sum(hmod)
        self.resolution = np.sqrt(energy_var)
        #print("Resolution:", self.resolution)

        # for consistency check per channel
        selection_model = self.x_select
        low_res_selected = selection_model.transform(low_res_data, keep_dictionary_structure=True)
        for channel in self.get_channels():
            B, P, _ = low_res_selected[channel].shape
            print(f"Calculate PCA on {channel}")
            low_pca = self.channel_pca[channel].fit_transform(low_res_selected[channel].reshape(B*P, -1))
            self.ood[channel].fit(low_pca)

        print("End of fit.")

        return high_res.reshape((B, P, -1))

    def check_compatibility_per_channel(self, low_res_data: Dict[str, np.ndarray], pulse_spacing: Optional[Dict[str, List[int]]]=None) -> Dict[str, np.ndarray]:
        """
        Check if a new low-resolution data source is compatible with the one used in training, by
        comparing the effect of the trained PCA model on it, but do it per channel.

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).
          pulse_spacing: The pulse spacing in multi-pulse data.

        Returns: Outlier score. If it is bigger than 0, this is an outlier.
        """
        selection_model = self.x_select
        channels = list(low_res_data.keys())
        # check if each channel is close to the mean
        low_res_selected = selection_model.transform(low_res_data, keep_dictionary_structure=True, pulse_spacing=pulse_spacing)

        def is_inlier(in_data, ch: str) -> np.ndarray:
            B, P, _ = in_data.shape
            data_pca = self.channel_pca[ch].transform(in_data.reshape((B*P, -1)))
            return self.ood[ch].predict(data_pca).reshape((B, P))
        result = {ch: is_inlier(low_res_selected[ch], ch) for ch in channels}
        return result

    def check_compatibility(self, low_res_data: Dict[str, np.ndarray], pulse_spacing: Optional[Dict[str, List[int]]]=None, pulse_energy: Optional[np.ndarray]=None) -> np.ndarray:
        """
        Check if a new low-resolution data source is compatible with the one used in training, by
        using a robust covariance matrix estimate of the data

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).
          pulse_spacing: The pulse spacing in multi-pulse data.
          beam_intensity: Beam intensity.

        Returns: An outlier score: if it is greater than 0, this is an outlier.
        """
        low_res = self.x_select.transform(low_res_data, pulse_spacing=pulse_spacing, pulse_energy=pulse_energy)
        B, P, _ = low_res.shape
        pca_model = self.x_model
        low_pca = pca_model.transform(low_res.reshape((B*P, -1)))
        return self.ood['full'].predict(low_pca).reshape((B, P))

    def xgm_profile(self) -> gaussian_kde:
        """Get KDE for the XGM intensity."""
        return self.kde_xgm

    def predict(self, low_res_data: Dict[str, np.ndarray], pulse_spacing: Optional[Dict[str, List[int]]]=None, pulse_energy: Optional[np.ndarray]=None) -> Dict[str, np.ndarray]:
        """
        Predict a high-resolution spectrum from a low resolution given one.
        The output includes the uncertainty in its second and third entries of the first dimension.

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).
          pulse_spacing: Where each pulse starts, relative to the first pulse (which is at 0). For single pulse data, set this to [0].

        Returns: High resolution data with shape (train_id, energy channel) in a dictionary containing
                 the expected prediction in key "expected", the stat. uncertainty in key "unc" and
                 a (1, energy channel) array for the PCA syst. uncertainty in key "pca".
        """
        #t = list()
        #n = list()
        #t += [time_ns()*1e-9]
        #n += ["Initial"]

        low_res_pre = self.x_select.transform(low_res_data, pulse_spacing=pulse_spacing, pulse_energy=pulse_energy)
        B, P, _ = low_res_pre.shape
        low_res_pre = low_res_pre.reshape((B*P, -1))
        #t += [time_ns()*1e-9]
        #n += ["Select"]

        low_pca = self.x_model.transform(low_res_pre)
        #t += [time_ns()*1e-9]
        #n += ["PCA x"]

        high_pca, high_pca_unc = self.fit_model.predict(low_pca, return_std=True)
        #t += [time_ns()*1e-9]
        #n += ["Fit model"]

        n_trains = high_pca.shape[0]
        # Note: The whiten=True setting in the PCA model leads to an affine transformation
        pca_y = np.concatenate((high_pca,
                                high_pca + high_pca_unc,
                               ),
                               axis=0)
        #t += [time_ns()*1e-9]
        #n += ["Concat"]

        high_res_predicted = self.y_model["pca"].inverse_transform(pca_y)
        expected = high_res_predicted[:n_trains, :]
        e_plus = high_res_predicted[n_trains:(2*n_trains), :]
        unc = np.fabs(e_plus - expected)
        pca_unc = self.y_model['unc'].uncertainty()
        total_unc = np.sqrt(pca_unc**2 + unc**2)

        #t += [time_ns()*1e-9]
        #n += ["Unc"]

        #t = np.diff(np.array(t))
        #n = n[1:]
        #print("Times")
        #print(dict(zip(n, t)))

        return dict(expected=expected.reshape((B, P, -1)),
                    unc=unc.reshape((B, P, -1)),
                    pca=pca_unc,
                    total_unc=total_unc.reshape((B, P, -1)),
                    )

    def deconvolve(self, expected: np.ndarray) -> np.ndarray:
        """
        Return Wiener filter deconvolved spectrum.

        Args:
          expected: The predicted spectrum.

        Returns: The Wiener filter-corrected spectrum.
        """
        W = self.wiener_filter_ft
        if len(expected.shape) == 2:
            W = W.reshape(1, -1)
        elif len(expected.shape) == 3:
            W = W.reshape(1, 1, -1)
        return np.real(np.absolute(np.fft.ifft(np.fft.fft(expected, axis=-1) * W, axis=-1)))

    def save(self, filename: str):
        """
        Save the fit model in a file.

        Args:
          filename: File name where to save this.
        """
        joblib.dump([self.x_select,
                     self.x_model,
                     self.y_model,
                     self.fit_model.state_dict() if self.bnn else self.fit_model,
                     self.channel_pca,
                     #self.channel_fit_model
                     DataHolder(dict(
                                     mu_xgm=self.mu_xgm,
                                     sigma_xgm=self.sigma_xgm,
                                     wiener_filter_ft=self.wiener_filter_ft,
                                     wiener_filter=self.wiener_filter,
                                     wiener_energy=self.wiener_energy,
                                     wiener_energy_ft=self.wiener_energy_ft,
                                     resolution=self.resolution,
                                     transfer_function=self.transfer_function,
                                     impulse_response=self.impulse_response,
                                     bnn=self.bnn,
                                    )
                               ),
                     self.ood,
                     self.kde_xgm,
                     ], filename, compress='zlib')

    @staticmethod
    def load(filename: str) -> Model:
        """
        Load model from a file.

        Args:
          filename: Name of the file where to read the model from.

        Returns: A new model object.
        """
        (x_select,
         x_model, y_model, fit_model,
         channel_pca,
         extra,
         ood,
         kde_xgm,
        ) = joblib.load(filename)

        obj = Model()

        extra = extra.get_data()
        obj.mu_xgm = extra["mu_xgm"]
        obj.sigma_xgm = extra["sigma_xgm"]
        obj.wiener_filter_ft = extra["wiener_filter_ft"]
        obj.wiener_filter = extra["wiener_filter"]
        obj.wiener_energy_ft = extra["wiener_energy_ft"]
        obj.wiener_energy = extra["wiener_energy"]
        obj.resolution = extra["resolution"]
        obj.transfer_function = extra["transfer_function"]
        obj.impulse_response = extra["impulse_response"]
        obj.bnn = extra["bnn"]

        obj.x_select = x_select
        obj.x_model = x_model
        obj.y_model = y_model
        if obj.bnn:
            obj.fit_model = BNNModel(state_dict=fit_model)
        else:
            obj.fit_model = fit_model
        obj.channel_pca = channel_pca
        #obj.channel_fit_model = channel_fit_model
        obj.ood = ood
        obj.kde_xgm = kde_xgm

        return obj

