#!/usr/bin/env python

import sys
sys.path.append('.')
sys.path.append('..')

import os
import argparse

import numpy as np
from extra_data import open_run, by_id, RunDirectory
from pes_to_spec.model import Model, matching_ids

from itertools import product

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1.inset_locator import InsetPosition
import seaborn as sns
import scipy
from scipy.signal import fftconvolve

from typing import Dict, Optional

from time import time_ns
import pandas as pd

SMALL_SIZE = 12
MEDIUM_SIZE = 18
BIGGER_SIZE = 22

plt.rc('font', size=BIGGER_SIZE)         # controls default text sizes
plt.rc('axes', titlesize=BIGGER_SIZE)    # fontsize of the axes title
plt.rc('axes', labelsize=BIGGER_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('ytick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('legend', fontsize=MEDIUM_SIZE)   # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

def plot_pes(filename: str, pes_raw_int: np.ndarray, first: int, last: int):
    """
    Plot low-resolution spectrum.

    Args:
      filename: Output file name.
      pes_raw_int: Low-resolution spectrum.

    """
    fig = plt.figure(figsize=(16, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.plot(np.arange(first, last), pes_raw_int, c='b', lw=3, label="Low-resolution measurement")
    ax.legend()
    ax.set(title=f"",
           xlabel="ToF index",
           ylabel="Intensity")
    fig.savefig(filename)
    plt.close(fig)

def plot_result(filename: str,
                spec_pred: Dict[str, np.ndarray],
                spec_smooth: np.ndarray,
                spec_raw_pe: np.ndarray,
                spec_raw_int: Optional[np.ndarray]=None,
                pes: Optional[np.ndarray]=None,
                pes_to_show: Optional[str]="",
                pes_bin: Optional[np.ndarray]=None,
                ):
    """
    Plot result with uncertainty band.

    Args:
      filename: Output file name.
      spec_pred: Predicted result with uncertainty bands in a dictionary.
      spec_smooth: Smoothened expected result with shape (features,).
      spec_raw_pe: x axis with the photon energy in eV.
      spec_raw_int: Original true expected result with shape (features,).
      pes: PES spectrum for the inset.
      pes_to_show: Name of the channel shown.
      pes_bin: PES bins.

    """
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    unc_stat = spec_pred["unc"]
    unc_pca = spec_pred["pca"]
    unc = np.sqrt(unc_stat**2 + unc_pca**2)
    ax.plot(spec_raw_pe, spec_smooth, c='b', lw=3, label="High-res. measurement (smoothened)")
    ax.plot(spec_raw_pe, spec_pred["expected"], c='r', ls='--', lw=3, label="High-res. prediction")
    #ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc, spec_pred["expected"] + unc, facecolor='green', alpha=0.6, label="68% unc.")
    ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc, spec_pred["expected"] + unc, facecolor='gold', alpha=0.5, label="68% unc.")
    #ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc_stat, spec_pred["expected"] + unc_stat, facecolor='red', alpha=0.6, label="68% unc. (stat.)")
    #ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc_pca, spec_pred["expected"] + unc_pca, facecolor='magenta', alpha=0.6, label="68% unc. (syst., PCA)")
    #if spec_raw_int is not None:
    #    ax.plot(spec_raw_pe, spec_raw_int, c='b', lw=1, ls='--', label="High-resolution measurement")
    #if wiener is not None:
    #    deconvolved = fftconvolve(spec_pred["expected"], wiener, mode="same")
    ax.plot(spec_raw_pe, spec_pred["deconvolved"], c='g', ls='-.', lw=3, label="Wiener filter result")
    Y = np.amax(spec_smooth)
    ax.legend(frameon=False, borderaxespad=0, loc='upper left')
    ax.set(title=f"", #avg(stat unc) = {unc_stat}, avg(pca unc) = {unc_pca}",
           xlabel="Photon energy [eV]",
           ylabel="Intensity",
           ylim=(0, 1.2*Y))
    if pes is not None:
        ax2 = plt.axes([0,0,1,1])
        # Manually set the position and relative size of the inset axes within ax1
        ip = InsetPosition(ax, [0.65,0.6,0.35,0.4])
        ax2.set_axes_locator(ip)
        Ypes = np.amax(pes)
        ax2.plot(pes_bin, pes, c='black', lw=3)
        ax2.set(title=f"Low-resolution example data",
                xlabel="Bin",
                ylabel=f"{pes_to_show}",
                ylim=(0, 1.2*Ypes),
                #labelsize=SMALL_SIZE,
                #xticklabels=dict(fontdict=dict(fontsize=SMALL_SIZE)),
                #yticklabels=dict(fontdict=dict(fontsize=SMALL_SIZE)),
                )
        ax2.title.set_size(SMALL_SIZE)
        ax2.xaxis.label.set_size(SMALL_SIZE)
        ax2.yaxis.label.set_size(SMALL_SIZE)
        ax2.tick_params(axis='both', which='major', labelsize=SMALL_SIZE)
    fig.savefig(filename)
    plt.close(fig)

def main():
    """
    Main entry point. Reads some data, trains and predicts.
    """
    parser = argparse.ArgumentParser(prog="offline_analysis", description="Test pes2spec doing an offline analysis of the data.")
    parser.add_argument('-p', '--proposal', type=int, metavar='INT', help='Proposal number', default=2828)
    parser.add_argument('-r', '--run', type=str, metavar='INT,INT,...', help='Run numbers, comma-separated.', default=206)
    parser.add_argument('-t', '--test-run', type=int, metavar='INT', help='Run to test', default=None)
    parser.add_argument('-m', '--model', type=str, metavar='FILENAME', default="", help='Model to load. If given, do not train a model and just do inference with this one.')
    parser.add_argument('-d', '--directory', type=str, metavar='DIRECTORY', default=".", help='Where to save the results.')
    parser.add_argument('-S', '--spec', type=str, metavar='NAME', default="SA3_XTD10_SPECT/MDL/SPECTROMETER_SQS_NAVITAR:output", help='SPEC name')
    parser.add_argument('-P', '--pes', type=str, metavar='NAME', default="SA3_XTD10_PES/ADC/1:network", help='PES name')
    parser.add_argument('-X', '--xgm', type=str, metavar='NAME', default="SA3_XTD10_XGM/XGM/DOOCS:output", help='XGM name')
    parser.add_argument('-o', '--offset', type=int, metavar='INT', default=0, help='Train ID offset')
    parser.add_argument('-c', '--xgm_cut', type=float, metavar='INTENSITY', default=500, help='XGM intensity threshold in uJ.')
    parser.add_argument('-e', '--bnn', action="store_true", default=False, help='Use BNN?')
    parser.add_argument('-w', '--weight', action="store_true", default=True, help='Whether to reweight data as a function of the pulse energy to make it invariant to that.')

    args = parser.parse_args()

    print("Opening run ...")
    runs = args.run.split(',')
    runs = [int(r) for r in runs]
    # get run
    run_list = [open_run(proposal=args.proposal, run=r) for r in runs]
    run = run_list[0]
    if len(run_list) > 1:
        run = run.union(*run_list[1:])

    run_test = run
    other_run_test = False
    if "test_run" in args and args.test_run is not None:
        other_run_test = True
        run_test = open_run(proposal=args.proposal, run=args.test_run)

    #run = RunDirectory("/gpfs/exfel/data/scratch/tmichela/data/r0206")

    # ----------------Used in the first tests-------------------------
    # get train IDs and match them, so we are sure to have information from all needed sources
    # in this example, there is an offset of -2 in the SPEC train ID, so correct for it
    spec_offset = -2
    spec_name = 'SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output'
    pes_name = 'SA3_XTD10_PES/ADC/1:network'

    spec_offset = 0
    spec_name = 'SA3_XTD10_SPECT/MDL/SPECTROMETER_SQS_NAVITAR:output'
    pes_name = 'SA3_XTD10_PES/ADC/1:network'
    # -------------------End of test setup ----------------------------

    spec_offset = args.offset
    spec_name = args.spec
    pes_name = args.pes
    xgm_name = args.xgm

    pes_tid = run[pes_name, "digitizers.trainId"].ndarray()
    xgm_tid = run[xgm_name, "data.trainId"].ndarray()

    if len(args.model) == 0:
        spec_tid = spec_offset + run[spec_name, "data.trainId"].ndarray()
        # these are the train ID intersection
        # this could have been done by a select call in the RunDirectory, but it would not correct for the spec_offset
        tids = matching_ids(spec_tid, pes_tid, xgm_tid)

        # read the spec photon energy and intensity
        spec_raw_pe = run[spec_name, "data.photonEnergy"].select_trains(by_id[tids - spec_offset]).ndarray()
        spec_raw_int = run[spec_name, "data.intensityDistribution"].select_trains(by_id[tids - spec_offset]).ndarray()

    else: # when doing inference, no need to load SPEC data
        tids = pes_tid

    # reserve part of it for the test stage
    train_tids = tids[:-10]
    if other_run_test:
        pes_tidt = run_test[pes_name, "digitizers.trainId"].ndarray()
        xgm_tidt = run_test[xgm_name, "data.trainId"].ndarray()
        spec_tidt = run_test[spec_name, "data.trainId"].ndarray()
        test_tids = matching_ids(spec_tidt, pes_tidt, xgm_tidt)
    else:
        test_tids = tids

    # read the PES data for each channel
    channels = [f"channel_{i}_{l}"
                for i, l in product(range(1, 5), ["A", "B", "C", "D"])]
    pes_raw = {ch: run[pes_name, f"digitizers.{ch}.raw.samples"].select_trains(by_id[tids]).ndarray()
               for ch in channels}
    pes_raw_t = {ch: run_test[pes_name, f"digitizers.{ch}.raw.samples"].select_trains(by_id[test_tids]).ndarray()
                   for ch in channels}

    # select test SPEC data
    spec_raw_pe_t = run_test[spec_name, "data.photonEnergy"].select_trains(by_id[test_tids - spec_offset]).ndarray()
    spec_raw_int_t = run_test[spec_name, "data.intensityDistribution"].select_trains(by_id[test_tids - spec_offset]).ndarray()

    print("Data in memory.")

    # read the XGM information
    #xgm_pressure = run['SA3_XTD10_XGM/XGM/DOOCS', "pressure.pressureFiltered.value"].select_trains(by_id[tids]).ndarray()
    #xgm_pe =  run['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[tids]).ndarray()
    #retvol_raw = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].select_trains(by_id[tids]).ndarray()
    #retvol_raw_timestamp = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.timestamp"].select_trains(by_id[tids]).ndarray()

    xgm_flux =  run['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[tids]).ndarray()[:, 0][:, np.newaxis]
    xgm_flux_t =  run_test['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[test_tids]).ndarray()[:, 0][:, np.newaxis]

    t = list()
    t_names = list()

    model = Model(bnn=args.bnn)

    train_idx = np.isin(tids, train_tids) & (xgm_flux[:,0] > args.xgm_cut)
    # we just need this for training and we need to avoid copying it, which blows up the memoray usage
    for k in pes_raw.keys():
        pes_raw[k] = pes_raw[k][train_idx]

    model.debug_peak_finding(pes_raw, os.path.join(args.directory, "test_peak_finding.png"))
    if len(args.model) == 0:
        print("Fitting")
        start = time_ns()
        w = model.uniformize(xgm_flux[train_idx])
        if not args.weight:
            w[...] = 1.0
        print("w", np.amin(w), np.amax(w), np.median(w))
        model.fit(pes_raw,
                  #{k: v[train_idx]
                  #for k, v in pes_raw.items()},
                   spec_raw_int[train_idx],
                   spec_raw_pe[train_idx],
                   w,
                   pulse_energy=xgm_flux[train_idx],
                   )
        t += [time_ns() - start]
        t_names += ["Fit"]

        print("Saving the model")
        start = time_ns()
        modelFilename = os.path.join(args.directory, "model.joblib")
        model.save(modelFilename)
        t += [time_ns() - start]
        t_names += ["Save"]
    else:
        print("Model has been given, so I will just load it.")
        modelFilename = args.model

    print("Loading the model")
    start = time_ns()
    model = Model.load(modelFilename)
    t += [time_ns() - start]
    t_names += ["Load"]

    # plot Wiener filter
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    plt.plot(np.fft.fftshift(model.wiener_energy_ft), np.fft.fftshift(np.absolute(model.wiener_filter_ft)))
    ax.set(title=f"",
           xlabel=r"Reciprocal energy [1/eV]",
           ylabel="Filter intensity [a.u.]",
           yscale='log',
           )
    fig.savefig(os.path.join(args.directory, "wiener_ft.png"))
    plt.close(fig)

    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    plt.plot(model.wiener_energy, np.absolute(model.wiener_filter))
    ax.set(title=f"",
           xlabel=r"Energy [eV]",
           ylabel="Filter value [a.u.]",
           )
    fig.savefig(os.path.join(args.directory, "wiener.png"))
    plt.close(fig)

    print("Check consistency")
    start = time_ns()
    Z = model.check_compatibility(pes_raw_t, pulse_energy=xgm_flux_t)
    print("Consistency check:", Z)
    Z = model.check_compatibility_per_channel(pes_raw_t)
    print("Consistency per channel:", Z)
    t += [time_ns() - start]
    t_names += ["Consistency"]

    # test
    print("Predict")
    start = time_ns()
    spec_pred = model.predict(pes_raw_t, pulse_energy=xgm_flux_t)
    spec_pred["deconvolved"] = model.deconvolve(spec_pred["expected"])
    t += [time_ns() - start]
    t_names += ["Predict"]

    print("Time taken in ms")
    df_time = pd.DataFrame(data=dict(time=t, name=t_names))
    df_time.time *= 1e-6
    print(df_time)

    print("Plotting")
    showSpec = False
    if len(args.model) == 0:
        showSpec = True
        spec_smooth = model.preprocess_high_res(spec_raw_int_t)

        # chi2 w.r.t XGM intensity
        de = (spec_raw_pe_t[0,1] - spec_raw_pe_t[0,0])
        chi2 = np.sum((spec_smooth[:, np.newaxis, :] - spec_pred["expected"])**2/(spec_pred["total_unc"]**2), axis=(-1, -2))
        ndof = float(spec_smooth.shape[1]) - 1.0
        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        ax.scatter(chi2/ndof, xgm_flux_t[:,0], c='r', s=20)
        ax.set(title=f"", #avg(stat unc) = {unc_stat}, avg(pca unc) = {unc_pca}",
               xlabel=r"$\chi^2/$ndof",
               ylabel="XGM intensity [uJ]",
               xlim=(0, 5),
               )
        ax2 = plt.axes([0,0,1,1])
        # Manually set the position and relative size of the inset axes within ax1
        ip = InsetPosition(ax, [0.65,0.6,0.35,0.4])
        ax2.set_axes_locator(ip)
        ax2.scatter(chi2/ndof, xgm_flux_t[:,0], c='r', s=30)
        #ax2.scatter(chi2/ndof, np.sum(spec_pred["expected"], axis=1)*de, c='b', s=30)
        #ax2.scatter(chi2/ndof, np.sum(spec_raw_int, axis=1)*de, c='g', s=30)
        ax2.set(title="",
                xlabel=r"$\chi^2/$ndof",
                ylabel=f"XGM intensity [uJ]",
                )
        ax2.title.set_size(SMALL_SIZE)
        ax2.xaxis.label.set_size(SMALL_SIZE)
        ax2.yaxis.label.set_size(SMALL_SIZE)
        ax2.tick_params(axis='both', which='major', labelsize=SMALL_SIZE)
        fig.savefig(os.path.join(args.directory, "intensity_vs_chi2.png"))
        plt.close(fig)

        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        sns.kdeplot(x=chi2/ndof, ax=ax)
        ax.set(title=f"",
               xlabel=r"$\chi^2/$ndof",
               ylabel="Density [a.u.]",
               xlim=(0, 5),
               )
        ax.text(0.90, 0.95, fr"$\mu = ${np.mean(chi2/ndof):.2f}",
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes,
                color='black', fontsize=15)
        ax.text(0.90, 0.90, fr"$\sigma = ${np.std(chi2/ndof):.2f}",
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes,
                color='black', fontsize=15)
        fig.savefig(os.path.join(args.directory, "chi2.png"))
        plt.close(fig)

        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        sns.kdeplot(x=xgm_flux_t[:,0], ax=ax)
        ax.set(title=f"",
               xlabel="XGM intensity [uJ]",
               ylabel="Density [a.u.]",
               )
        ax.text(0.90, 0.95, fr"$\mu = ${np.mean(xgm_flux_t[:,0]):.2f}",
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes,
                color='black', fontsize=15)
        ax.text(0.90, 0.90, fr"$\sigma = ${np.std(xgm_flux_t[:,0]):.2f}",
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes,
                color='black', fontsize=15)
        fig.savefig(os.path.join(args.directory, "intensity.png"))
        plt.close(fig)

        # rmse
        rmse = np.sqrt(np.mean((spec_smooth[:, np.newaxis, :] - spec_pred["expected"])**2, axis=(-1, -2)))
        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        ax.scatter(rmse, xgm_flux_t[:,0], c='r', s=30)
        ax = plt.gca()
        ax.set(title=f"",
               xlabel=r"Root-mean-squared error",
               ylabel="XGM intensity [uJ]",
               )
        fig.savefig(os.path.join(args.directory, "intensity_vs_rmse.png"))
        plt.close(fig)

        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        sns.kdeplot(x=rmse, ax=ax)
        ax.set(title=f"",
               xlabel="Root-mean-squared error",
               ylabel="Density [a.u.]",
               xlim=(0, 20),
               )
        ax.text(0.90, 0.95, fr"$\mu = ${np.mean(rmse):.2f}",
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes,
                color='black', fontsize=15)
        ax.text(0.90, 0.90, fr"$\sigma = ${np.std(rmse):.2f}",
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes,
                color='black', fontsize=15)
        fig.savefig(os.path.join(args.directory, "rmse.png"))
        plt.close(fig)

        # SPEC integral w.r.t XGM intensity
        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        sns.regplot(x=np.sum(spec_raw_int_t, axis=1)*de, y=xgm_flux_t[:,0], color='r', robust=True, ax=ax)
        ax.set(title=f"",
               xlabel="SPEC (raw) integral",
               ylabel="XGM Intensity [uJ]",
               )
        fig.savefig(os.path.join(args.directory, "xgm_vs_intensity.png"))
        plt.close(fig)

        # SPEC integral w.r.t XGM intensity
        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        sns.regplot(x=np.sum(spec_raw_int_t, axis=-1)*de, y=np.sum(spec_pred["expected"], axis=(-1, -2))*de, color='r', robust=True, ax=ax)
        ax.set(title=f"",
               xlabel="SPEC (raw) integral",
               ylabel="Predicted integral",
               )
        fig.savefig(os.path.join(args.directory, "expected_vs_intensity.png"))
        plt.close(fig)

        fig = plt.figure(figsize=(12, 8))
        gs = GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        sns.regplot(x=np.sum(spec_pred["expected"], axis=(-1, -2))*de, y=xgm_flux_t[:,0], color='r', robust=True, ax=ax)
        ax.set(title=f"",
               xlabel="Predicted integral",
               ylabel="XGM intensity [uJ]",
               )
        fig.savefig(os.path.join(args.directory, "xgm_vs_expected.png"))
        plt.close(fig)

    first, last = model.get_low_resolution_range()
    first += 10
    last -= 10
    pes_to_show = 'channel_1_D'
    # plot
    for idx in range(len(spec_raw_int_t) - 10, len(spec_raw_int_t)):
        tid = test_tids[idx]
        plot_result(os.path.join(args.directory, f"test_{tid}.png"),
                   {k: item[idx, 0, ...] if k != "pca"
                       else item[0, ...]
                       for k, item in spec_pred.items()},
                    spec_smooth[idx, :] if showSpec else None,
                    spec_raw_pe_t[idx, :] if showSpec else None,
                    spec_raw_int_t[idx, :] if showSpec else None,
                    )
        for ch in channels:
            plot_pes(os.path.join(args.directory, f"test_pes_{tid}_{ch}.png"),
                     pes_raw_t[ch][idx, first:last], first, last)

if __name__ == '__main__':
    main()

